
App.Medicine.Surgery = {};
/**
 * Composes the Procedure object from its parts
 *
 * This function has the only purpose to ensure the result object has all required properties
 * @param {string} typeId
 * @param {string} label
 * @param {number} effect
 * @param {string} desc
 * @param {slaveOperation} [action]
 * @param {number} [costs] money costs
 * @param {number} [hCosts] health costs
 * @param {string} [surgeryType]
 * @returns {App.Medicine.Surgery.Procedure}
 */
App.Medicine.Surgery.makeOption = function(typeId, label, effect, desc, action, costs, hCosts, surgeryType) {
	return {
		typeId: typeId,
		label: label,
		targetEffect: effect,
		description: desc,
		costs: costs,
		healthCosts: hCosts,
		action: action,
		surgeryType: surgeryType
	};
};

/**
 * Composes the procedure option with empty .action, i.e. a procedure that can't be applied
 * @param {string} typeId
 * @param {string} label
 * @param {string} desc
 * @returns {App.Medicine.Surgery.Procedure}
 */
App.Medicine.Surgery.makeImpossibleOption = function(typeId, label, desc) {
	return this.makeOption(typeId, label, 0, desc);
};

/**
 * Various constants for procedures
 */
App.Medicine.Keys = {
	Surgery: {
		Target: {
			/**
			 * Type id's for breast-related procedures
			 */
			breast: {
				installImplant: "breast.implant.install",
				removeImplant: "breast.implant.remove",
				fillUp: "breast.implant.fill",
				drain: "breast.implant.drain",
				changeImplant: "breast.implant.replace",
				reduction: "breast.tissue.reduce",
			},
			/**
			 * Type id's for butt-related procedures
			 */
			butt: {
				installImplant: "butt.implant.install",
				removeImplant: "butt.implant.remove",
				fillUp: "butt.implant.fill",
				drain: "butt.implant.drain",
				changeImplant: "but.implant.replace",
				reduction: "butt.tissue.reduce",
			}
		}
	}
};

/**
 * Commit procedure, executing its action and subtracting its costs
 * @param {App.Medicine.Surgery.Procedure} surgery
 * @param {App.Entity.SlaveState} slave
 */
App.Medicine.Surgery.commit = function(surgery, slave) {
	const V = State.variables;
	V.surgeryType = surgery.surgeryType;
	surgery.action(slave);
	cashX(forceNeg(surgery.costs), "slaveSurgery", slave);
	slave.health -= (V.PC.medicine >= 100) ? Math.round(surgery.healthCosts / 2) : surgery.healthCosts;
};

/**
 * Returns markup for a link to execute the given procedure
 * @param {string} passage Passage to go after the surgery
 * @param {App.Medicine.Surgery.Procedure} surgery
 * @param {App.Entity.SlaveState} slave
 * @returns {string}
 */
App.Medicine.Surgery.makeLink = function(passage, surgery, slave) {
	if (surgery.action === undefined) {
		return App.UI.disabledLink(surgery.label, [surgery.description]);
	}

	function healthCosts() {
		const hc = (State.variables.PC.medicine >= 100) ? Math.round(surgery.healthCosts / 2) : surgery.healthCosts;
		if (hc > 30) {
			return 'substantial';
		} else if (hc > 20) {
			return 'significant';
		} else if (hc > 10) {
			return 'moderate';
		} else if (hc > 5) {
			return 'light';
		}
		return 'insignificant';
	}

	return App.UI.link(surgery.label, App.Medicine.Surgery.commit, [surgery, slave], passage,
		`${capFirstChar(surgery.description)}.\nSurgery costs: ${cashFormat(surgery.costs)}.\nProjected health damage: ${healthCosts()}.`);
};

/**
 * Helpers for composing procedure descriptions
 */
App.Medicine.Surgery.ListHelpers = class {
	/**
	 * @param {App.Entity.SlaveState} slave
	 * @param {string} bodyPart
	 * @param {Object.<string, string>} keys
	 * @param {App.Utils.Pronouns} pronouns
	 * @param {boolean} showCCs
	 */
	constructor(slave, bodyPart, keys, pronouns, showCCs) {
		/** @private */
		this._slave = slave;
		/** @private */
		this._bodyPart = bodyPart;
		/** @private */
		this._keys = keys;
		/** @private */
		this._pronouns = pronouns;
		/** @private */
		this._V = State.variables;
		this._showCCs = showCCs;
	}
	/**
	 * @param {string} name
	 * @param {number} implantType
	 * @param {number} size
	 * @returns {App.Medicine.Surgery.Procedure}
	 */
	installImplants(name, implantType, size) {
		return App.Medicine.Surgery.makeOption(this._keys.installImplant, `${capFirstChar(name)} implants`, size,
			`place ${name}${this._showCCs ? ` ${size}cc` : ''} implants into ${this._pronouns.his} ${this._bodyPart}`,
			slave => {
				slave[`${this._bodyPart}Implant`] = size;
				slave[`${this._bodyPart}ImplantType`] = implantType;
				slave[this._bodyPart] += size;
			}, this._V.surgeryCost, 10, this._bodyPart
		);
	}

	removeImplants() {
		return App.Medicine.Surgery.makeOption(this._keys.removeImplant, "Remove implants",
			-this._slave[`${this._bodyPart}Implant`],
			`remove ${this._pronouns.his} ${this._bodyPart} implants`,
			slave => {
				slave[`${this._bodyPart}`] -= slave[`${this._bodyPart}Implant`];
				slave[`${this._bodyPart}Implant`] = 0;
				slave[`${this._bodyPart}ImplantType`] = 0;
			}, this._V.surgeryCost, 5, `${this._bodyPart}Loss`
		);
	}

	/**
	 * @param {string} name
	 * @param {number} implantType
	 * @param {number} size
	 * @returns {App.Medicine.Surgery.Procedure}
	 */
	replaceImplants(name, implantType, size) {
		return App.Medicine.Surgery.makeOption(this._keys.changeImplant, `${capFirstChar(name)} implants`,
			size - this._slave.boobsImplant,
			`replace ${this._pronouns.his} ${this._bodyPart} implants with ${name}${this._showCCs ? ` (${size}cc)` : ''} ones`,
			slave => {
				slave[this._bodyPart] += size - slave[`${this._bodyPart}Implant`];
				slave[`${this._bodyPart}Implant`] = size;
				slave[`${this._bodyPart}ImplantType`] = implantType;
			}, this._V.surgeryCost, 10, this._bodyPart
		);
	}

	/**
	 * @param {number} volume
	 * @returns {App.Medicine.Surgery.Procedure}
	 */
	fillUp(volume) {
		return App.Medicine.Surgery.makeOption(this._keys.fillUp, "Add inert filler", volume,
			`add ${this._showCCs ? `${volume}cc of` : 'some'} inert filler to each of ${this._pronouns.his} ${this._bodyPart} implants`,
			slave => {
				slave[`${this._bodyPart}Implant`] += volume;
				slave[this._bodyPart] += volume;
			},
			this._V.surgeryCost, 10, this._bodyPart
		);
	}

	/**
	 * @param {number} volume
	 * @returns {App.Medicine.Surgery.Procedure}
	 */
	drain(volume) {
		return App.Medicine.Surgery.makeOption(this._keys.drain, `Drain ${volume}cc`, -volume,
			`drain ${this._showCCs ? `${volume}cc of` : 'some'} inert filler from ${this._pronouns.his} ${this._bodyPart} implants`,
			slave => {
				slave[`${this._bodyPart}Implant`] -= volume;
				slave[this._bodyPart] -= volume;
			}, this._V.surgeryCost, 5, `${this._bodyPart}Loss`
		);
	}

	/**
	 * @param {string} procedureName
	 * @param {number} sizeChange
	 * @returns {App.Medicine.Surgery.Procedure}
	 */
	reduce(procedureName, sizeChange) {
		return App.Medicine.Surgery.makeOption(this._keys.reduction,
			`${capFirstChar(procedureName)} ${this._bodyPart}`, -200,
			`${procedureName} ${this._pronouns.his} ${this._bodyPart}`,
			slave => {
				slave[this._bodyPart] -= sizeChange;
			}, this._V.surgeryCost, 5, `${this._bodyPart}Loss`
		);
	}
};

/**
 * Returns options to accept all possible surgeries
 * @returns {App.Medicine.Surgery.SizingOptions}
 */
App.Medicine.Surgery.allSizingOptions = function() {
	return {
		augmentation: true,
		reduction: true,
		strings: true,
		replace: true
	};
};

App.Medicine.Surgery.sizingProcedures = function() {
	return {
		bodyPart: bodyPart,
		boobs: boobSizingProcedures,
		butt: buttSizingProcedures
	};

	/**
	 * Returns list of available surgeries targeted at changing size of the given body part
	 * @param {string} bodyPart
	 * @param {App.Entity.SlaveState} slave
	 * @param {App.Medicine.Surgery.SizingOptions} [options]
	 * @returns {App.Medicine.Surgery.Procedure[]}
	 */
	function bodyPart(bodyPart, slave, options) {
		switch (bodyPart) {
			case "boob":
			case "boobs":
			case "breast":
			case "breasts":
				return boobSizingProcedures(slave, options);
			case "ass":
			case "booty":
			case "butt":
				return buttSizingProcedures(slave, options);
			default:
				throw `No sizing procedures for ${bodyPart}`;
		}
	}

	/**
	 * @param {App.Entity.SlaveState} slave
	 * @param {App.Medicine.Surgery.SizingOptions} [options]
	 * @returns {App.Medicine.Surgery.Procedure[]}
	 */
	function boobSizingProcedures(slave, options = {}) {
		const V = State.variables;
		const thisArcology = V.arcologies[0];
		const largeImplantsAvailable = thisArcology.FSTransformationFetishistResearch === 1;
		const advancedFillableImplantsAvailable = V.ImplantProductionUpgrade === 1;
		const pronouns = getPronouns(slave);
		const {he, His} = pronouns;

		const types = App.Medicine.Keys.Surgery.Target.breast; // shortcuts
		const helper = new App.Medicine.Surgery.ListHelpers(slave, "boobs", types, pronouns, V.showBoobCCs);

		let r = [];
		if (options.augmentation) {
			if (slave.boobs > 50000) {
				r.push(App.Medicine.Surgery.makeImpossibleOption(types.fillUp, "Increase boobs", `<em>${His} breasts are as large as ${he} can physically support</em>`));
			} else if (slave.boobsImplant > 10000 && !largeImplantsAvailable && slave.boobsImplantType !== 1) {
				r.push(App.Medicine.Surgery.makeImpossibleOption(types.fillUp, "Increase boobs", `<em>${His} implants are filled to capacity</em>`));
			} else if (slave.indentureRestrictions >= 2) {
				r.push(App.Medicine.Surgery.makeImpossibleOption(types.installImplant, "Change boob size", `<em>${His} indenture forbids elective surgery</em>`));
			} else if (slave.breastMesh === 1) {
				r.push(App.Medicine.Surgery.makeImpossibleOption(types.installImplant, "Put implants", `<em>${His} supportive mesh implant blocks implantation</em>`));
			} else if (slave.boobsImplant === 0) {
				if (options.strings) {
					r.push(helper.installImplants("string", 1, 400));
				}
				if (V.surgeryUpgrade === 1) {
					r.push(helper.installImplants("large", 0, 600));
				}
				r.push(helper.installImplants("standard", 0, 400));
				r.push(helper.installImplants("small", 0, 200));
			} else if (slave.boobsImplant > 10000 && slave.boobsImplantType !== 1 && advancedFillableImplantsAvailable) {
				r.push(helper.fillUp(1000));
			} else if (slave.boobsImplant > 9800 && slave.boobsImplantType !== 1 && advancedFillableImplantsAvailable) {
				r.push(helper.replaceImplants("hyper fillable", 0, 11000));
			} else if (slave.boobsImplant > 2000 && slave.boobsImplantType !== 1) {
				r.push(helper.fillUp(400));
			} else if (slave.boobsImplant > 1800 && slave.boobsImplantType !== 1) {
				r.push(App.Medicine.Surgery.makeImpossibleOption(types.fillUp, "Add inert filler", `<em>${His} implants are filled to capacity</em>`));

				const advancedFillable = helper.replaceImplants("advanced fillable", 0, 2200);
				if (!advancedFillableImplantsAvailable) {
					advancedFillable.costs += 10000;
					advancedFillable.label += " (special order)";
					advancedFillable.description += " (special order)";
				}
				r.push(advancedFillable);
			} else if (slave.boobsImplant > 600 && slave.boobsImplantType !== 1) {
				r.push(helper.fillUp(200));
			} else if (slave.boobsImplant > 400 && slave.boobsImplantType !== 1) {
				r.push(helper.replaceImplants("fillable", 0, 800));
			} else if (slave.boobsImplant > 200 && slave.boobsImplantType !== 1) {
				r.push(helper.replaceImplants("large", 0, 600));
			} else if (slave.boobsImplant > 0 && slave.boobsImplantType !== 1) {
				r.push(helper.replaceImplants("standard", 0, 400));
				r.push(helper.replaceImplants("large", 0, 600));
			}
		}

		if (options.reduction && (slave.boobs > 300 || slave.boobsImplant > 0)) {
			if (slave.boobsImplant > 0) {
				if (slave.boobsImplantType === 1 && slave.boobsImplant > 400) {
					if (slave.boobsImplant > 8000) {
						r.push(helper.drain(1000));
					} else if (slave.boobsImplant > 5000) {
						r.push(helper.drain(750));
					} else if (slave.boobsImplant > 2000) {
						r.push(helper.drain(500));
					} else if (slave.boobsImplant > 1000) {
						r.push(helper.drain(250));
					} else if (slave.boobsImplant > 500) {
						r.push(helper.drain(100));
					}
				}
				r.push(helper.removeImplants());
			}
			if ((slave.boobs > 300) && (slave.boobsImplant === 0) && slave.indentureRestrictions < 2) {
				r.push(helper.reduce("reduce", 200));
				if (slave.boobs < 675) {
					r.push(helper.reduce("slightly reduce", 25));
				}
			}
			if ((slave.boobsImplant === 0) && slave.indentureRestrictions < 2 && (slave.breedingMark !== 1 || V.propOutcome !== 1 || V.eugenicsFullControl === 1 || V.arcologies[0].FSRestart === "unset")) {
				if (slave.boobs >= 7000) {
					r.push(App.Medicine.Surgery.makeOption(types.reduction, "Mastectomy", 300 - slave.boobs,
						"perform mastectomy",
						slave => {
							slave.boobs = 300;
						},
						V.surgeryCost, 30, "mastectomy+"
					));
				} else if (slave.boobs >= 2000) {
					r.push(App.Medicine.Surgery.makeOption(types.reduction, "Mastectomy", 300 - slave.boobs,
						"perform mastectomy",
						slave => {
							slave.boobs = 300;
						},
						V.surgeryCost, 30, "mastectomy"
					));
				}
			}
		}
		return r;
	}

	/**
	 * @param {App.Entity.SlaveState} slave
	 * @param {App.Medicine.Surgery.SizingOptions} [options]
	 * @returns {App.Medicine.Surgery.Procedure[]}
	 */
	function buttSizingProcedures(slave, options = {}) {
		const V = State.variables;
		const thisArcology = V.arcologies[0];
		const largeImplantsAvailable = thisArcology.FSTransformationFetishistResearch === 1;
		const advancedFillableImplantsAvailable = V.ImplantProductionUpgrade === 1;
		const pronouns = getPronouns(slave);
		const {he, His} = pronouns;

		const types = App.Medicine.Keys.Surgery.Target.butt; // shortcuts
		const helper = new App.Medicine.Surgery.ListHelpers(slave, "butt", types, pronouns, false);

		let r = [];

		if (options.augmentation) {
			if (slave.indentureRestrictions >= 2) {
				r.push(App.Medicine.Surgery.makeImpossibleOption(types.installImplant, "Change butt size", `<em>${His} indenture forbids elective surgery</em>`));
			} else if (slave.butt > 19) {
				r.push(App.Medicine.Surgery.makeImpossibleOption(types.fillUp, "Increase butt", `<em>${His} butt is as large as it can possibly get</em>`));
			} else if (slave.butt > 7 && !largeImplantsAvailable && thisArcology.FSAssetExpansionist === "unset" && slave.buttImplantType !== 1) {
				r.push(App.Medicine.Surgery.makeImpossibleOption(types.installImplant, "Increase butt", `<em>${His} butt is as large as ${he} can physically support</em>`));
			} else if (slave.buttImplant > 7 && !largeImplantsAvailable && slave.buttImplantType !== 1) {
				r.push(App.Medicine.Surgery.makeImpossibleOption(types.installImplant, "Increase butt", `<em>${His} butt implants are filled to capacity</em>`));
			} else if (slave.buttImplant === 0) {
				r.push(helper.installImplants("standard", 0, 1));
				if (options.strings) {
					r.push(helper.installImplants("string", 1, 1));
				}
			} else if (slave.buttImplant === 1 && slave.buttImplantType !== 1) {
				r.push(helper.replaceImplants("bigger", 0, 2));
			} else if (slave.buttImplant === 2 && slave.buttImplantType !== 1) {
				r.push(helper.replaceImplants("fillable", 0, 3));
			} else if (slave.buttImplant === 4 && slave.buttImplantType !== 1) {
				r.push(App.Medicine.Surgery.makeImpossibleOption(types.fillUp, "Increase size", `<em>${His} implants are filled to capacity</em>`));
				const advancedFillable = helper.replaceImplants("advanced fillable", 0, 5);
				if (!advancedFillableImplantsAvailable) {
					advancedFillable.costs += 10000;
					advancedFillable.label += " (special order)";
					advancedFillable.description += " (special order)";
				}
				r.push(advancedFillable);
			} else if (slave.buttImplant === 8 && slave.buttImplantType !== 1 && advancedFillableImplantsAvailable) {
				r.push(helper.replaceImplants("hyper fillable", 0, 9));
			} else if (slave.buttImplant > 2 && slave.buttImplantType !== 1) {
				r.push(helper.fillUp(1));
			}
		}

		if (options.reduction) {
			if (slave.buttImplant > 0) {
				if (slave.indentureRestrictions < 2) {
					if (slave.buttImplantType === 1 && slave.buttImplant > 1) {
						r.push(helper.drain(1));
					}
					r.push(helper.removeImplants());
				}
			}
			if ((slave.butt > 1) && (slave.buttImplant === 0)) {
				if (slave.indentureRestrictions < 2) {
					r.push(helper.reduce("reduce", 1));
				}
			}
		}
		return r;
	}
}();

/**
* Clean up extremities on removal or piercings, tats, and brands
* @param {App.Entity.SlaveState} slave
* @param {string} part
*/

window.surgeryAmp = function(slave, part) {
	if (part === "left ear") {
		slave.earShape = "none";
		slave.earT = "none";
		delete slave.brand["left ear"];
		slave.earPiercing = 0;
		slave.health -= 10;
	} else if (part === "right ear") {
		slave.earShape = "none";
		slave.earT = "none";
		delete slave.brand["right ear"];
		// slave.earPiercing = 0;
		slave.health -= 10;
	} else if (part === "left arm") {
		delete slave.brand["left upper arm"];
		delete slave.brand["left lower arm"];
		delete slave.brand["left wrist"];
		delete slave.brand["left hand"];
		// slave.armsTat = 0;
		// slave.nails = 0;
		slave.armAccessory = "none";
		slave.health -= 10;
	} else if (part === "right arm") {
		delete slave.brand["right upper arm"];
		delete slave.brand["right lower arm"];
		delete slave.brand["right wrist"];
		delete slave.brand["right hand"];
		// slave.armsTat = 0;
		// slave.nails = 0;
		slave.armAccessory = "none";
		slave.health -= 10;
	} else if (part === "left leg") {
		delete slave.brand["left thigh"];
		delete slave.brand["left calf"];
		delete slave.brand["left ankle"];
		delete slave.brand["left foot"];
		slave.legsTat = 0;
		slave.shoes = "none";
		slave.legAccessory = "none";
		slave.heightImplant = 0;
		slave.health -= 10;
	} else if (part === "right leg") {
		delete slave.brand["right thigh"];
		delete slave.brand["right calf"];
		delete slave.brand["right ankle"];
		delete slave.brand["right foot"];
		slave.legsTat = 0;
		slave.shoes = "none";
		slave.legAccessory = "none";
		slave.heightImplant = 0;
		slave.health -= 10;
	} else if (part === "dick") {
		slave.dick = 0;
		slave.foreskin = 0;
		slave.skill.vaginal = 0;
		slave.dickPiercing = 0;
		slave.dickTat = 0;
		slave.dickAccessory = "none";
		slave.chastityPenis = 0;
		slave.health -= 20;
	} else if (part === "vagina") {
		slave.vagina = -1;
		slave.ovaries = 0;
		slave.preg = -2;
		slave.pregSource = 0;
		slave.skill.vaginal = 0;
		slave.vaginaTat = 0;
		slave.vaginalAccessory = "none";
		slave.vaginalAttachment = "none";
		slave.chastityVagina = 0;
		slave.health -= 20;
	} else if (part === "horn") {
		slave.horn = "none";
		slave.hornColor = "none";
		slave.health -= 10;
	} else if (part === "voicebox") {
		slave.voice = 0;
		slave.voiceImplant = 0;
		slave.health -= 10;
	} else if (part === "lips") {
		slave.lips -= slave.lipsImplant;
		slave.lipsImplant = 0;
		if (slave.skill.oral > 10) {
			slave.skill.oral -= 10;
		}
	}
};

/**
 * Prepare and set up for new Fuckdoll
 * @param {App.Entity.SlaveState} slave
 */
window.beginFuckdoll = function(slave) {
	slave.fuckdoll = 1;
	slave.toyHole = "all her holes";
	if ((slave.pubicHStyle !== "bald") && (slave.pubicHStyle !== "hairless")) {
		slave.pubicHStyle = "waxed";
	}
	slave.livingRules = "spare";
	slave.speechRules = "restrictive";
	slave.releaseRules = "restrictive";
	slave.relationshipRules = "restrictive";
	slave.choosesOwnClothes = 0;
	slave.clothes = "a Fuckdoll suit";
	slave.collar = "none";
	if ((slave.missingLegs !== 3) || (slave.shoes !== "none")) {
		slave.shoes = "heels";
	}
	slave.armAccessory = "none";
	slave.legAccessory = "none";
	slave.vaginalAccessory = "none";
	slave.vaginalAttachment = "none";
	slave.dickAccessory = "none";
	slave.buttplug = "none";
	slave.chastityAnus = 0;
	slave.chastityPenis = 0;
	slave.chastityVagina = 0;
	slave.attrKnown = 1;
	slave.fetishKnown = 1;
	slave.subTarget = 0;
	slave.sentence = 0;
	slave.training = 0;
	slave.inflation = 0;
	slave.inflationType = "none";
	slave.inflationMethod = 0;
	slave.milkSource = 0;
	slave.cumSource = 0;
};
